// -*- C++ -*-
/*!
 * @file CorbaNaming.h
 * @brief CORBA naming service helper class
 * @date $Date: 2007/04/13 15:35:21 $
 * @author Noriaki Ando <n-ando@aist.go.jp>
 *
 * Copyright (C) 2006
 *     Noriaki Ando
 *     Task-intelligence Research Group,
 *     Intelligent Systems Research Institute,
 *     National Institute of
 *         Advanced Industrial Science and Technology (AIST), Japan
 *     All rights reserved.
 *
 * $Id: CorbaNaming.h,v 1.2 2007/04/13 15:35:21 n-ando Exp $
 *
 */

/*
 * $Log: CorbaNaming.h,v $
 * Revision 1.2  2007/04/13 15:35:21  n-ando
 * Error handing processing in case NameServer does not exist was added.
 * Some bug fixes.
 *
 * Revision 1.1  2006/11/04 19:43:01  n-ando
 * CORBA Naming service helper class has rewritten and renamed.
 *
 */


#ifndef CorbaNaming_h
#define CorbaNaming_h

#include <rtm/RTC.h>
#ifdef ORB_IS_OMNIORB
#undef PACKAGE_BUGREPORT
#undef PACKAGE_NAME
#undef PACKAGE_STRING
#undef PACKAGE_TARNAME
#undef PACKAGE_VERSION
#include <omniORB4/CORBA.h>
#endif

// STL includes
#include <map>
#include <string>
#include <vector>

namespace RTC
{
  /*!
   * @if jp
   * @class CorbaNaming
   * @brief CORBA Naming Service wp[NX
   *
   * ̃NX́ACosNaming::NamingContext ɑ΂郉bp[NXłB
   * CosNaming::NamingContext Iy[VƂقړ@\
   * Iy[V񋟂ƂƂɁAl[R|[lg CosNaming::Name
   * ̑ɕɂ閼O\󂯕tIy[V񋟂B
   *
   * IuWFNg͐A邢͐ CORBA l[T[oɐڑ
   * ȌÃl[T[õ[gReLXgɑ΂ĎX̃Iy[V
   * B
   * [Kw̃l[~OReLXg̍쐬IuWFNg̃oChɂāA
   * r̃ReLXg݂ȂꍇłAIɃReLXgoCh
   * ړĨReLXgIuWFNg̃oChsƂłB
   *
   * @else
   * @class CorbaNaming
   * @brief CORBA Naming Service helper class
   *
   * This class is a wrapper class of CosNaming::NamingContext.
   * Almost the same operations which CosNaming::NamingContext has are
   * provided, and some operation allows string naming representation of
   * context and object instead of CosNaming::Name.
   *
   * The object of the class would connect to a CORBA naming server at
   * the instantiation or immediately after instantiation.
   * After that the object invokes operations to the root context of it.
   * This class realizes forced binding to deep NamingContext, without binding
   * intermediate NamingContexts explicitly.
   *
   * @endif
   */
  class CorbaNaming
  {
  public:
    CorbaNaming(CORBA::ORB_ptr orb);
    CorbaNaming(CORBA::ORB_ptr orb, const char* name_server);
    virtual ~CorbaNaming(){};
    void init(const char* name_server);

    typedef CosNaming::NamingContext::NotFound      NotFound;
    typedef CosNaming::NamingContext::CannotProceed CannotProceed;
    typedef CosNaming::NamingContext::InvalidName   InvalidName;
    typedef CosNaming::NamingContext::AlreadyBound  AlreadyBound;
    typedef CosNaming::NamingContext::NotEmpty      NotEmpty;
    typedef CosNaming::NamingContextExt::InvalidAddress InvalidAddress;
    typedef std::vector<CORBA::Object_ptr> ObjectList;


    /*!
     * @if jp
     *
     * @brief Object  bind 
     *
     * CosNaming::bind() Ƃقړ̓邪Aɗ^ꂽl[T[o
     * [gReLXgɑ΂bind()Ăяo_قȂB
     *
     * Name <name>  Object <obj> 𓖊Y NamingContext ɃoChB
     * c_n  n Ԗڂ NameComponent 킷ƂƁA
     * name  n  NameComponent 琬ƂAȉ̂悤ɈB
     *
     * cxt->bind(<c_1, c_2, ... c_n>, obj) ͈ȉ̑ƓłB
     * cxt->resolve(<c_1, ... c_(n-1)>)->bind(<c_n>, obj)
     *
     * Ȃ킿A1Ԗڂn-1Ԗڂ̃ReLXgAn-1Ԗڂ̃ReLXg
     *  name <n> Ƃā@obj  bind B
     * OɎQ <c_1, ... c_(n-1)>  NemingContext ́A
     * bindContext()  rebindContext() ŊɃoChς݂łȂ΂ȂȂB
     *  <c_1, ... c_(n-1)>  NamingContext ݂Ȃꍇɂ́A
     * NotFound OB
     *
     * AoChtO force  true ̎́A<c_1, ... c_(n-1)>
     * ݂ȂꍇɂAċAIɃReLXgoChȂA
     * ŏII obj 𖼑O name <c_n> ɃoChB
     *
     * ̏ꍇłAn-1Ԗڂ̃ReLXg name<n> ̃IuWFNg
     * (Object 邢 ReLXg) oChĂ
     * AlreadyBound OB
     *
     * @param name IuWFNgɕt閼O NameComponent
     * @param obj ֘At Object
     * @param force truȅꍇAr̃ReLXgIɃoCh
     *
     * @exception NotFound r <c_1, c_2, ..., c_(n-1)> ݂ȂB
     * @exception CannotProceed 炩̗RŏpłȂB
     * @exception InvalidName  name ̖OsB
     * @exception AlreadyBound name <c_n>  Object łɃoChĂB
     *
     * @else
     *
     * @brief
     *
     * @endif
     */
    void bind(const CosNaming::Name& name, CORBA::Object_ptr obj,
	      const bool force = 1)
      throw(NotFound, CannotProceed, InvalidName, AlreadyBound);


    /*!
     * @if jp
     *
     * @brief Object  bind 
     *
     * Object  bind ۂɗ^閼O\ł邱ƈȊÓAbind()
     * ƓłBbind(toName(string_name), obj) ƓB
     *
     * @param string_name IuWFNgɕt閼O̕\
     * @param obj ֘AtIuWFNg
     * @param force truȅꍇAr̃ReLXgIɃoCh
     *
     * @exception NotFound r <c_1, c_2, ..., c_(n-1)> ݂ȂB
     * @exception CannotProceed 炩̗RŏpłȂB
     * @exception InvalidName  name ̖OsB
     * @exception AlreadyBound name <n>  Object łɃoChĂB
     *
     * @else
     *
     * @brief
     *
     * @endif
     */
    void bindByString(const char* string_name, CORBA::Object_ptr obj,
		      const bool force = 1)
      throw(NotFound, CannotProceed, InvalidName, AlreadyBound);

    /*!
     * @if jp
     *
     * @brief r̃ReLXg bind Ȃ Object  bind 
     *
     * context ŗ^ꂽ NamingContext ɑ΂āAname Ŏw肳ꂽ
     * l[R|[lg <c_1, ... c_(n-1)>  NamingContext Ƃ
     * ȂAO <c_n> ɑ΂ obj  bind B
     * A<c_1, ... c_(n-1)> ɑΉ NamingContext Ȃꍇɂ
     * V NamingContext oChB
     *
     * ŏII <c_1, c_2, ..., c_(n-1)> ɑΉ NamingContext 
     * ܂͉ꂽŁACosNaming::bind(<c_n>, object) ĂяoB
     * ̂ƂAłɃoCfBO݂ AlreadyBoundOB
     *
     * r̃ReLXgߒŁA悤ƂReLXg
     * O NamingContext ł͂Ȃ Binding ݂ꍇA
     * CannotProceed O𒆎~B
     *
     * @param context bind Jn@NamingContext
     * @param name IuWFNgɕt閼Õl[R|[lg
     * @param obj ֘AtIuWFNg
     *
     * @exception CannotProceed <c_1, ..., c_(n-1)> ɑΉ NamingContext 
     *            ̂ЂƂAł NamingContext ȊO object ɃoCh
     *            ĂApłȂB
     * @exception InvalidName O name s
     * @exception AlreadyBound name <c_n> ɂłɉ炩 object oCh
     *            ĂB
     * @else
     *
     * @brief
     *
     * @endif
     */
    void bindRecursive(CosNaming::NamingContext_ptr context,
		       const CosNaming::Name& name,
		       CORBA::Object_ptr obj)
      throw(CannotProceed, InvalidName, AlreadyBound);



    /*!
     * @if jp
     *
     * @brief Object  rebind 
     *
     * name Ŏw肳ꂽ Binding łɑ݂ꍇ bind() Ɠ
     * łBoCfBOłɑ݂ꍇɂ́AVoCfBO
     * uB
     *
     * @param name IuWFNgɕt閼O NameComponent
     * @param obj ֘AtIuWFNg
     * @param force truȅꍇAr̃ReLXgIɃoCh
     *
     * @else
     *
     * @brief
     *
     * @endif
     */
    void rebind(const CosNaming::Name& name, CORBA::Object_ptr obj,
		const bool force = 1)
      throw(NotFound, CannotProceed, InvalidName);


    /*!
     * @if jp
     *
     * @brief Object  rebind 
     *
     * Object  rebind ۂɗ^閼O\ł邱ƈȊO rebind()
     * ƓłBrebind(toName(string_name), obj) ƓB
     *
     * @param string_name IuWFNgɕt閼O̕\
     * @param obj ֘AtIuWFNg
     * @param force truȅꍇAr̃ReLXgIɃoCh
     *
     * @exception NotFound r <c_1, c_2, ..., c_(n-1)> ݂ȂB
     * @exception CannotProceed 炩̗RŏpłȂB
     * @exception InvalidName  name ̖OsB
     *
     * @else
     *
     * @brief
     *
     * @endif
     */
    void rebindByString(const char* string_name, CORBA::Object_ptr obj,
			const bool force = 1)
      throw(NotFound, CannotProceed, InvalidName);

    
    /*!
     * @if jp
     *
     * @brief r̃ReLXg bind Ȃ Object  rebind 
     *
     * name <c_n> Ŏw肳ꂽ NamingContext  Object łɑ݂
     * ꍇ bindRecursive() ƓłB
     *
     * name <c_n> Ŏw肳ꂽoCfBOłɑ݂ꍇɂ́A
     * VoCfBOɒuB
     *
     * @param name IuWFNgɕt閼O̕\
     * @param obj ֘AtIuWFNg
     * @param force truȅꍇAr̃ReLXgIɃoCh
     *
     * @exception CannotProceed r̃ReLXgłȂB
     * @exception InvalidName ^ꂽ name sB
     *
     * @else
     *
     * @brief
     *
     * @endif
     */
    void rebindRecursive(CosNaming::NamingContext_ptr context,
			 const CosNaming::Name& name,
			 CORBA::Object_ptr obj)
      throw(CannotProceed, InvalidName);

    /*!
     * @if jp
     *
     * @brief NamingContext  bind 
     *
     * bind IuWFNg NamingContext ł邱Ƃ bind() 
     * ƓłB
     *
     * @param name IuWFNgɕt閼Õl[R|[lg
     * @param name_cxt ֘At NamingContext
     * @param force truȅꍇAr̃ReLXgIɃoCh
     *
     * @else
     *
     * @brief
     *
     * @endif
     */
    void bindContext(const CosNaming::Name& name,
		     CosNaming::NamingContext_ptr name_cxt,
		     const bool force = 1)
      throw(NotFound, CannotProceed, InvalidName, AlreadyBound);

    /*!
     * @if jp
     *
     * @brief NamingContext  bind 
     *
     * bind IuWFNg NamingContext ł邱Ƃ bind() 
     * ƓłB
     *
     * @param name IuWFNgɕt閼O̕\
     * @param name_cxt ֘At NamingContext
     * @param force truȅꍇAr̃ReLXgIɃoCh
     *
     * @else
     *
     * @brief
     *
     * @endif
     */
    void bindContext(const char* string_name,
		     CosNaming::NamingContext_ptr name_cxt,
		     const bool force = 1)
      throw(NotFound, CannotProceed, InvalidName, AlreadyBound);

    /*!
     * @if jp
     *
     * @brief NamingContext  bind 
     *
     * bind IuWFNg NamingContext ł邱Ƃ
     * bindRecursive() ƓłB
     *
     * @param context bind Jn@NamingContext
     * @param name IuWFNgɕt閼Õl[R|[lg
     * @param name_cxt ֘At NamingContext
     * @param force truȅꍇAr̃ReLXgIɃoCh
     *
     * @else
     *
     * @brief
     *
     * @endif
     */
    void bindContextRecursive(CosNaming::NamingContext_ptr context,
			      const CosNaming::Name& name,
			      CosNaming::NamingContext_ptr name_cxt);
    /*!
     * @if jp
     *
     * @brief NamingContext  rebind 
     *
     * name Ŏw肳ꂽReLXgłɑ݂ꍇ bindContext() 
     * ƓłB
     * oCfBOłɑ݂ꍇɂ́AVoCfBO
     * uB
     *
     * @param name IuWFNgɕt閼Õl[R|[lg
     * @param name_cxt ֘At NamingContext
     * @param force truȅꍇAr̃ReLXgIɃoCh
     *
     * @else
     *
     * @brief
     *
     * @endif
     */
    void rebindContext(const CosNaming::Name& name,
		       CosNaming::NamingContext_ptr name_cxt,
		       const bool force = 1)
      throw(NotFound, CannotProceed, InvalidName);

    /*!
     * @if jp
     *
     * @brief NamingContext  rebind 
     *
     * name Ŏw肳ꂽReLXgłɑ݂ꍇ bindContext() 
     * ƓłB
     * oCfBOłɑ݂ꍇɂ́AVoCfBO
     * uB
     *
     * @param name IuWFNgɕt閼O̕\
     * @param name_cxt ֘At NamingContext
     * @param force truȅꍇAr̃ReLXgIɃoCh
     *
     * @else
     *
     * @brief
     *
     * @endif
     */
    void rebindContext(const char* string_name,
		       CosNaming::NamingContext_ptr name_cxt,
		       const bool force = 1)
      throw(NotFound, CannotProceed, InvalidName);


    void rebindContextRecursive(CosNaming::NamingContext_ptr context,
				const CosNaming::Name& name,
				CosNaming::NamingContext_ptr name_cxt);

    /*!
     * @if jp
     *
     * @brief Object  name 
     *
     * name  bind ĂIuWFNgQƂԂB
     * l[R|[lg <c_1, c_2, ... c_n> ͍ċAIɉB
     * 
     * CosNaming::resolve() Ƃقړ̓邪Aɗ^ꂽ
     * l[T[õ[gReLXgɑ΂ resolve() Ăяo_
     * قȂB
     *
     * @param name ׂIuWFNg̖Õl[R|[lg
     * @return ꂽIuWFNgQ
     *
     * @else
     *
     * @endif
     */
    CORBA::Object_ptr resolve(const CosNaming::Name& name)
      throw(NotFound, CannotProceed, InvalidName);

    /*!
     * @if jp
     *
     * @brief Object  name 
     *
     * name  bind ĂIuWFNgQƂԂB
     * l[R|[lg <c_1, c_2, ... c_n> ͍ċAIɉB
     * 
     * CosNaming::resolve() Ƃقړ̓邪Aɗ^ꂽ
     * l[T[õ[gReLXgɑ΂ resolve() Ăяo_
     * قȂB
     *
     * @param name ׂIuWFNg̖O̕\
     * @return ꂽIuWFNgQ
     *
     * @else
     *
     * @endif
     */
    CORBA::Object_ptr resolve(const char* string_name)
      throw(NotFound, CannotProceed, InvalidName);

    /*!
     * @if jp
     *
     * @brief w肳ꂽÕIuWFNg bind 
     *
     * name  bind ĂIuWFNgQƂԂB
     * l[R|[lg <c_1, c_2, ... c_n> ͍ċAIɉB
     * 
     * CosNaming::unbind() Ƃقړ̓邪Aɗ^ꂽ
     * l[T[õ[gReLXgɑ΂ unbind() Ăяo_
     * قȂB
     *
     * @param name ׂIuWFNg̖Õl[R|[lg
     * @return ꂽIuWFNgQ
     *
     * @else
     *
     * @endif
     */
    void unbind(const CosNaming::Name& name)
      throw(NotFound, CannotProceed, InvalidName);

    /*!
     * @if jp
     *
     * @brief w肳ꂽÕIuWFNg bind 
     *
     * name  bind ĂIuWFNgQƂԂB
     * l[R|[lg <c_1, c_2, ... c_n> ͍ċAIɉB
     * 
     * CosNaming::unbind() Ƃقړ̓邪Aɗ^ꂽ
     * l[T[õ[gReLXgɑ΂ unbind() Ăяo_
     * قȂB
     *
     * @param name ׂIuWFNg̖O̕\
     * @return ꂽIuWFNgQ
     *
     * @else
     *
     * @endif
     */
    void unbind(const char* string_name)
      throw(NotFound, CannotProceed, InvalidName);

    /*!
     * @if jp
     *
     * @brief VReLXg𐶐
     *
     * ^ꂽl[T[oŐꂽ NamingContext ԂB
     * Ԃꂽ NamingContext  bind ĂȂB
     * 
     * @return ꂽV NamingContext
     *
     * @else
     *
     * @endif
     */
    CosNaming::NamingContext_ptr newContext();

    /*!
     * @if jp
     *
     * @brief VReLXg bind 
     *
     * ^ꂽ name ɑ΂ĐVReLXgoChB
     * ꂽ@NamingContext ̓l[T[oŐꂽ̂łB
     * 
     * @param name NamingContextɕt閼Õl[R|[lg
     * @return ꂽV NamingContext
     *
     * @else
     *
     * @endif
     */
    CosNaming::NamingContext_ptr
    bindNewContext(const CosNaming::Name& name, bool force = true)
      throw(NotFound, CannotProceed, InvalidName, AlreadyBound);

    /*!
     * @if jp
     *
     * @brief VReLXg bind 
     *
     * ^ꂽ name ɑ΂ĐVReLXgoChB
     * ꂽ@NamingContext ̓l[T[oŐꂽ̂łB
     * 
     * @param name NamingContextɕt閼O̕\
     * @return ꂽV NamingContext
     *
     * @else
     *
     * @endif
     */
    CosNaming::NamingContext_ptr
    bindNewContext(const char* string_name, bool force = true)
      throw(NotFound, CannotProceed, InvalidName, AlreadyBound);

    /*!
     * @if jp
     *
     * @brief NamingContext ANeBu
     *
     * context Ŏw肳ꂽ NamingContext ANeBuB
     * context ɑ̃ReLXgoChĂꍇ NotEmpty O
     * B
     * 
     * @param context ANeBu NamingContext
     *
     * @else
     *
     * @brief Destroy the naming context
     *
     * Delete the specified naming context.
     * any bindings should be <unbind> in which the given context is bound to
     * some names before invoking <destroy> operation on it. 
     *
     * @param context NamingContext which is destroied.
     *     
     * @endif
     */
    void destroy(CosNaming::NamingContext_ptr context)
      throw(NotEmpty);


    /*!
     * @if jp
     * @brief NamingContext ċAIɉĔANeBu
     * @else
     * @brief Destroy the naming context recursively
     * @endif
     */
    void destroyRecursive(CosNaming::NamingContext_ptr context)
      throw(NotEmpty, NotFound, CannotProceed, InvalidName);


    /*!
     * @if jp
     * @brief ׂĂ Binding 폜
     * @else
     * @brief Destroy all binding
     * @endif
     */
    void clearAll();


    /*!
     * @if jp
     * @brief ^ꂽ NamingContext  Binding 擾
     * @else
     * @brief Get Binding on the NamingContextDestroy all binding
     * @endif
     */
    void list(CosNaming::NamingContext_ptr name_cxt,
	      unsigned long how_many,
	      CosNaming::BindingList_var& bl,
	      CosNaming::BindingIterator_var& bi);


    //============================================================
    // interface of NamingContextExt
    //============================================================
    /*!
     * @if jp
     * @brief ^ꂽ NameComponent ̕\Ԃ
     * @else
     * @brief Get string representation of given NameComponent
     * @endif
     */
    char* toString(const CosNaming::Name& name)
      throw(InvalidName);
    

    /*!
     * @if jp
     * @brief ^ꂽ\ NameComponent ɕ
     * @else
     * @brief Get NameComponent from gien string name representation
     * @endif
     */
    CosNaming::Name toName(const char* string_name)
      throw(InvalidName);


    /*!
     * @if jp
     * @brief ^ꂽ addre  string_name  URL\擾
     * @else
     * @brief Get URL representation from given addr and string_name
     * @endif
     */
    char* toUrl(char* addr, char* string_name)
      throw(InvalidAddress, InvalidName);


    /*!
     * @if jp
     * @brief ^ꂽ\ resolve IuWFNgԂ
     * @else
     * @brief Resolve from name of string representation and get object 
     * @endif
     */
    CORBA::Object_ptr resolveStr(const char* string_name)
      throw(NotFound, CannotProceed, InvalidName, AlreadyBound);


    //============================================================
    // Find functions
    //============================================================

    //    ObjectList find(const char* name, const char* kind);
    //    ObjectList findById(const char* name, const char* kind);
    //    ObjectList findByKind(const char* name, const char* kind);

    /*!
     * @if jp
     * @brief OoCh܂͉
     * @else
     * @brief Bind of resolve the given name component
     * @endif
     */
    CORBA::Object_ptr bindOrResolve(CosNaming::NamingContext_ptr context,
				    const CosNaming::Name& name,
				    CORBA::Object_ptr obj);
    
    /*!
     * @if jp
     * @brief OoCh܂͉
     * @else
     * @brief Bind of resolve the given name component
     * @endif
     */
    CosNaming::NamingContext_ptr
    bindOrResolveContext(CosNaming::NamingContext_ptr context,
			 const CosNaming::Name& name,
			 CosNaming::NamingContext_ptr new_context);

    
    /*!
     * @if jp
     * @brief OoCh܂͉
     * @else
     * @brief Bind of resolve the given name component
     * @endif
     */
    CosNaming::NamingContext_ptr
    bindOrResolveContext(CosNaming::NamingContext_ptr context,
			 const CosNaming::Name& name);


    /*!
     * @if jp
     * @brief l[T[o̖O擾
     * @else
     * @brief Get the name of naming server
     * @endif
     */
    const char* getNameServer();
    
    /*!
     * @if jp
     * @brief [gReLXg擾
     * @else
     * @brief Get the root context
     * @endif
     */
    CosNaming::NamingContext_ptr getRootContext();

    
    /*!
     * @if jp
     * @brief IuWFNgl[~OReLXgʂ
     * @else
     * @brief Whether the object is NamingContext
     * @endif
     */
    bool isNamingContext(CORBA::Object_ptr obj);
    
    /*!
     * @if jp
     * @brief ^ꂽOl[~OReLXgǂ
     * @else
     * @brief Whether the given name component is NamingContext
     * @endif
     */
    bool isNamingContext(const CosNaming::Name& name);
    
    /*!
     * @if jp
     * @brief ^ꂽOl[~OReLXgǂ
     * @else
     * @brief Whether the given string name is NamingContext
     * @endif
     */
    bool isNamingContext(const char* string_name);
    
    /*!
     * @if jp
     * @brief l[R|[lg̕Ԃ
     * @else
     * @brief Get subset of given name component
     * @endif
     */
    CosNaming::Name subName(const CosNaming::Name& name,
			    long begin,
			    long end = -1);


  protected:
    /*!
     * @if jp
     * @brief l[R|[lg̕\擾
     * @else
     * @brief Get string representation of name component
     * @endif
     */
    void nameToString(const CosNaming::Name& name, char* string_name,
		      unsigned long slen);
    /*!
     * @if jp
     * @brief l[R|[lg̕\̕擾
     * @else
     * @brief Get string length of the name component's string representation
     * @endif
     */
    CORBA::ULong getNameLength(const CosNaming::Name& name);

    /*!
     * @if jp
     * @brief ̕
     * @else
     * @brief Split of string
     * @endif
     */
    unsigned int split(const std::string& input,
		       const std::string& delimiter,
		       std::vector<std::string>& results);


    
    CORBA::ORB_var m_varORB;
    std::string m_nameServer;
    CosNaming::NamingContextExt_var m_rootContext;

  private:
    CORBA::ULong m_blLength;
    
  }; // class CorbaNaming

}; // namespace RTC

#endif // end of __Naming_h__
